# Copyright 2019 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for tfx.utils.channel."""

import tensorflow as tf
from tfx.dsl.placeholder import placeholder
from tfx.types import artifact
from tfx.types import channel

from ml_metadata.proto import metadata_store_pb2


class _MyType(artifact.Artifact):
  TYPE_NAME = 'MyTypeName'
  PROPERTIES = {
      'string_value': artifact.Property(artifact.PropertyType.STRING),
  }


class _AnotherType(artifact.Artifact):
  TYPE_NAME = 'AnotherTypeName'


class ChannelTest(tf.test.TestCase):

  def testValidChannel(self):
    instance_a = _MyType()
    instance_b = _MyType()
    chnl = channel.Channel(_MyType).set_artifacts([instance_a, instance_b])
    self.assertEqual(chnl.type_name, 'MyTypeName')
    self.assertCountEqual(chnl.get(), [instance_a, instance_b])

  def testInvalidChannelType(self):
    instance_a = _MyType()
    instance_b = _MyType()
    with self.assertRaises(ValueError):
      channel.Channel(_AnotherType).set_artifacts([instance_a, instance_b])

  def testStringTypeNameNotAllowed(self):
    with self.assertRaises(ValueError):
      channel.Channel('StringTypeName')

  def testJsonRoundTrip(self):
    chnl = channel.Channel(
        type=_MyType,
        additional_properties={
            'string_value': metadata_store_pb2.Value(string_value='forty-two')
        },
        additional_custom_properties={
            'int_value': metadata_store_pb2.Value(int_value=42)
        })
    serialized = chnl.to_json_dict()
    rehydrated = channel.Channel.from_json_dict(serialized)
    self.assertIs(chnl.type, rehydrated.type)
    self.assertEqual(chnl.type_name, rehydrated.type_name)
    self.assertEqual(chnl.additional_properties,
                     rehydrated.additional_properties)
    self.assertEqual(chnl.additional_custom_properties,
                     rehydrated.additional_custom_properties)

  def testJsonRoundTripUnknownArtifactClass(self):
    chnl = channel.Channel(type=_MyType)

    serialized = chnl.to_json_dict()
    serialized['type']['name'] = 'UnknownTypeName'

    rehydrated = channel.Channel.from_json_dict(serialized)
    self.assertEqual('UnknownTypeName', rehydrated.type_name)
    self.assertEqual(chnl.type._get_artifact_type().properties,
                     rehydrated.type._get_artifact_type().properties)
    self.assertTrue(rehydrated.type._AUTOGENERATED)

  def testFutureProducesPlaceholder(self):
    chnl = channel.Channel(type=_MyType)
    future = chnl.future()
    self.assertIsInstance(future, placeholder.ChannelWrappedPlaceholder)
    self.assertIs(future.channel, chnl)
    self.assertIsInstance(future[0], placeholder.ChannelWrappedPlaceholder)
    self.assertIsInstance(future.value, placeholder.ChannelWrappedPlaceholder)

  def testValidUnionChannel(self):
    channel1 = channel.Channel(type=_MyType)
    channel2 = channel.Channel(type=_MyType)
    union_channel = channel.union([channel1, channel2])
    self.assertIs(union_channel.type_name, 'MyTypeName')
    self.assertEqual(union_channel.channels, [channel1, channel2])

    union_channel = channel.union([channel1, channel.union([channel2])])
    self.assertIs(union_channel.type_name, 'MyTypeName')
    self.assertEqual(union_channel.channels, [channel1, channel2])

  def testMismatchedUnionChannelType(self):
    chnl = channel.Channel(type=_MyType)
    another_channel = channel.Channel(type=_AnotherType)
    with self.assertRaises(TypeError):
      channel.union([chnl, another_channel])

  def testEmptyUnionChannel(self):
    with self.assertRaises(AssertionError):
      channel.union([])


if __name__ == '__main__':
  tf.test.main()
